/** \file server/parse.h
 */

/* This file is part of LCDd, the lcdproc server.
 *
 * This file is released under the GNU General Public License.
 * Refer to the COPYING file distributed with this package.
 *
 * Copyright (c) 1999, William Ferrell, Selene Scriven
 */

#ifndef PARSE_H
#define PARSE_H

// This should be pretty self-explanatory...
int parse_all_client_messages(void);

#endif
